/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.battlefield;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.swing.event.EventListenerList;

import fb.battlefieldelements.Element;
import fb.battlefieldelements.ElementComparator;
import fb.battlefieldelements.Figure;
import fb.battlefieldelements.MagicEffect;
import fb.battlefieldelements.Projectile;
import fb.battlefieldelements.Sign;
import fb.battlefieldelements.VerticalSet;
import fb.battlefieldelements.Warrior;

/**
 * the battlefield data model
 * 
 * @author farfpeuf
 * 
 */
public class BattleFieldModel {

	// singleton
	private static BattleFieldModel instance;

	private List
	// figures on the battlefields
			figures = new ArrayList(),
			// warriors on the battlefields
			warriors = new ArrayList(),
			// elements on the battlefields
			elements = new ArrayList(),
			// signs on the battlefields
			signs = new ArrayList(),
			// projectiles on the battlefields
			projectiles = new ArrayList(),
			// magic effects on the battlefields
			magics = new ArrayList();

	// comparator use to sort the elements in the order in which they must be
	// drawn
	private ElementComparator comparator = new ElementComparator();

	/**
	 * sort the battlefields elements in the order in which they must be drawn
	 * 
	 * 
	 */
	public void sort() {
		Collections.sort(elements, comparator);
	}

	/**
	 * get singleton instance
	 * 
	 * @return singleton
	 */
	public static BattleFieldModel getInstance() {
		if (instance == null) {
			instance = new BattleFieldModel();
		}
		return instance;
	}

	/**
	 * add a sign to draw on the battlefield
	 * 
	 * @param sign
	 *            the new sign
	 */
	public void addSign(Sign sign) {
		signs.add(sign);
		fireBFChanged(new BattleFieldModelEvent(sign,
				BattleFieldModelEvent.INSERT));
	}

	/**
	 * remove a sign from the battlefield
	 * 
	 * @param sign
	 *            the sign to remove
	 */
	public void removeSign(Sign sign) {
		signs.remove(sign);
		fireBFChanged(new BattleFieldModelEvent(sign,
				BattleFieldModelEvent.DELETE));
	}

	/**
	 * add a magic effect to draw on the battlefield
	 * 
	 * @param effect
	 *            the magic effect
	 */
	public void addMagic(MagicEffect effect) {
		magics.add(effect);
		fireBFChanged(new BattleFieldModelEvent(effect,
				BattleFieldModelEvent.INSERT));
	}

	/**
	 * add a collection of magic effects
	 * 
	 * @param effects
	 *            the magic effects
	 */
	public void addMagics(Collection effects) {
		MagicEffect effect;
		Iterator iter = effects.iterator();
		while (iter.hasNext()) {
			effect = (MagicEffect) iter.next();
			addMagic(effect);
		}
	}

	/**
	 * remove a magic effect from the battlefield
	 * 
	 * @param effect
	 *            the magic effect to remove
	 */
	public void removeMagic(MagicEffect effect) {
		magics.remove(effect);
		fireBFChanged(new BattleFieldModelEvent(effect,
				BattleFieldModelEvent.DELETE));
	}

	/**
	 * get the number of projectiles on the battlefield
	 * 
	 * @return the number of projectiles
	 */
	public int getNbProjectiles() {
		return projectiles.size();
	}

	/**
	 * add a projectile to draw on the battlefield
	 * 
	 * @param projectile
	 *            the new projectile
	 */
	public void addProjectile(Projectile projectile) {
		projectiles.add(projectile);
		fireBFChanged(new BattleFieldModelEvent(projectile,
				BattleFieldModelEvent.INSERT));
	}

	/**
	 * remove a projectile from the battlefield
	 * 
	 * @param projectile
	 *            the projectile to remove
	 */
	public void removeProjectile(Projectile projectile) {
		projectiles.remove(projectile);
		fireBFChanged(new BattleFieldModelEvent(projectile,
				BattleFieldModelEvent.DELETE));
	}

	/**
	 * add a vertical set to draw on the battlefield
	 * 
	 * @param set
	 *            the vertical set
	 */
	public void addVerticalSet(VerticalSet set) {
		addElement(set);
	}

	/**
	 * add a warrior to draw on the battlefield
	 * 
	 * @param warrior
	 *            the new warrior
	 */
	public void addWarrior(Warrior warrior) {
		warriors.add(warrior);
		addFigure(warrior);
	}

	/**
	 * remove a warrior from the battlefield
	 * 
	 * @param warrior
	 *            the projectile to warrior
	 */
	public void removeWarrior(Warrior warrior) {
		warriors.remove(warrior);
		removeFigure(warrior);
	}

	/**
	 * add a figure to draw on the battlefield
	 * 
	 * @param figure
	 *            the new warrior
	 */
	public void addFigure(Figure anim) {
		figures.add(anim);
		addElement(anim);
	}

	/**
	 * remove a figure from the battlefield
	 * 
	 * @param figure
	 *            the figure to remove
	 */
	public void removeFigure(Figure figure) {
		figures.remove(figure);
		removeElement(figure);
	}

	/**
	 * add an element to draw on the battlefield
	 * 
	 * @param element
	 *            the new element
	 */
	private void addElement(Element element) {
		elements.add(element);
		fireBFChanged(new BattleFieldModelEvent(element,
				BattleFieldModelEvent.INSERT));
	}

	/**
	 * remove an element from the battlefield
	 * 
	 * @param element
	 *            the element to remove
	 */
	private void removeElement(Element element) {
		elements.remove(element);
		fireBFChanged(new BattleFieldModelEvent(element,
				BattleFieldModelEvent.DELETE));
	}

	/**
	 * get a warrior id
	 * 
	 * @param warrior
	 *            the warrior
	 * @return the warrior id
	 */
	public int getId(Warrior warrior) {
		return warriors.indexOf(warrior);
	}

	/**
	 * get the number of warriors on the battlefield
	 * 
	 * @return the number of warriors
	 */
	public int getNbWarriors() {
		return warriors.size();
	}

	/**
	 * get a warrior by specifying his id
	 * @param id the id
	 * @return the warrior
	 */
	public Warrior getWarrior(int id) {
		return (Warrior) warriors.get(id);
	}

	/**
	 * clear the model
	 *
	 */
	public void clear() {
		figures.clear();
		warriors.clear();
		elements.clear();
		signs.clear();
		projectiles.clear();
		magics.clear();
	}

	/**
	 * BattleFieldModelListener list
	 */
	protected EventListenerList listenerList = new EventListenerList();

	/**
	 * add a listener on the battlefields elements
	 * @param l the new listener
	 */
	public void addBFModelListener(BattleFieldModelListener l) {
		listenerList.add(BattleFieldModelListener.class, l);
	}

	/**
	 * remove a listener
	 * @param l the listener to remove
	 */
	public void removeBFModelListener(BattleFieldModelListener l) {
		listenerList.remove(BattleFieldModelListener.class, l);
	}

	/**
	 * notify listeners that the battlefield has changed
	 * @param e
	 */
	public void fireBFChanged(BattleFieldModelEvent e) {

		Object[] listeners = listenerList.getListenerList();

		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == BattleFieldModelListener.class) {
				((BattleFieldModelListener) listeners[i + 1]).bfChanged(e);
			}
		}
	}

	/**
	 * @return Returns the elements.
	 */
	public List getElements() {
		return elements;
	}

	/**
	 * @return Returns the figures.
	 */
	public List getFigures() {
		return figures;
	}

	/**
	 * @return Returns the magics.
	 */
	public List getMagics() {
		return magics;
	}

	/**
	 * @return Returns the projectiles.
	 */
	public List getProjectiles() {
		return projectiles;
	}

	/**
	 * @return Returns the signs.
	 */
	public List getSigns() {
		return signs;
	}

	/**
	 * @return Returns the warriors.
	 */
	public List getWarriors() {
		return warriors;
	}


}