/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.battlefieldelements;

import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

import fb.army.entities.SpellCaster;
import fb.battlefield.BattleFieldModel;
import fb.battlefield.SpriteSequence;
import fb.battlefield.PaintedWithSprite;
import fb.main.FB;
import fb.sound.SoundPlayer;
import fb.util.DX;
import fb.util.FBColors;
import fb.util.GeometryTools;

/**
 * An animated figure on the battlefield
 * 
 * @author farfpeuf
 * 
 */
public class Figure extends Animated implements Cloneable {

	// current sprite frame index
	protected int frameIndex;

	// the figue base area
	protected Area baseArea;

	// the current game coordinates
	protected int x = 0, y = 0;

	// the destination coordinates
	protected int xd, yd;

	// the orientation angle
	protected double angle = 0, tmpAngle;

	// the current sprite sequence
	private SpriteSequence currentSeq;

	// sprite animation counter
	protected int animationCounter = DX.roll(4);

	// current action
	protected char action = 'n';

	// is figure currently visible
	protected boolean visible = false;

	// PaintedWithSprite delegate
	protected PaintedWithSprite spriteDelegate;

	// list of all prepared projectiles
	private List projectiles;

	/**
	 * constructor
	 * 
	 * @param modelled
	 *            sprite delegate
	 */
	public Figure(PaintedWithSprite modelled) {
		super();
		this.spriteDelegate = modelled;
		projectiles = new ArrayList();

	}

	/**
	 * prepare a projectile before shooting it
	 * 
	 * @param projectile
	 *            the projectile
	 */
	public void prepareProjectile(Projectile projectile) {
		projectiles.add(projectile);
	}

	/**
	 * get the figure width
	 * 
	 * @return the width
	 */
	protected int width() {
		return (int) (spriteDelegate.getWidth() * 8);
	}

	/**
	 * get the figure length
	 * 
	 * @return the figure length
	 */
	protected int length() {
		return (int) (spriteDelegate.getLength() * 8);
	}

	/**
	 * get x offset
	 * 
	 * @return x offset
	 */
	protected int modelDx() {
		return (int) (0 - width() / 2);
	}

	/**
	 * get y offset
	 * 
	 * @return y offset
	 */
	protected int modelDy() {
		return (int) (0 - length() / 2);
	}

	/**
	 * get figure bounds
	 */
	public Rectangle getBounds() {
		return getBounds(spriteDelegate.getScale());
	}

	/**
	 * initialize baseArea and sprite sequence
	 * 
	 */
	public void init() {
		baseArea = new Area(new Rectangle(modelDx(), modelDy(), width(),
				length()));

		attribseq();
	}

	/**
	 * reset base area
	 * 
	 */
	public void resetAreas() {
		// create
		baseArea = new Area(new Rectangle(modelDx(), modelDy(), width(),
				length()));

		// move
		baseArea = new Area(AffineTransform.getTranslateInstance((double) xd,
				(double) yd).createTransformedShape(baseArea));

		// rotate
		baseArea = new Area(AffineTransform.getRotateInstance(-angle,
				(double) xd, (double) yd).createTransformedShape(baseArea));

	}

	/**
	 * teleport the figure
	 * 
	 */
	public void teleport() {
		doMove(xd - x, yd - y);
		visible = true;
	}

	/**
	 * clone the figure
	 */
	public Object clone() {
		try {
			return super.clone();
		} catch (CloneNotSupportedException cnse) {
			cnse.printStackTrace();
			return null;
		}
	}

	/**
	 * get x game coordinate
	 */
	public int getX() {
		return x;
	}

	/**
	 * get y game coordinate
	 */
	public int getY() {
		return y;
	}

	/**
	 * get base A point
	 * 
	 * @return base A point
	 */
	public Point getA() {
		Point p = new Point(xd + modelDx(), yd + modelDy());
		p.translate(-xd, -yd);
		AffineTransform.getRotateInstance(-angle).deltaTransform(p, p);
		p.translate(xd, yd);
		return p;
	}

	/**
	 * get base B point
	 * 
	 * @return base B point
	 */
	public Point getB() {
		Point p = new Point(xd - modelDx(), yd + modelDy());
		p.translate(-xd, -yd);
		AffineTransform.getRotateInstance(-angle).deltaTransform(p, p);
		p.translate(xd, yd);
		return p;
	}

	/**
	 * get base C point
	 * 
	 * @return base C point
	 */
	public Point getC() {
		Point p = new Point(xd - modelDx(), yd - modelDy());
		p.translate(-xd, -yd);
		AffineTransform.getRotateInstance(-angle).deltaTransform(p, p);
		p.translate(xd, yd);
		return p;
	}

	/**
	 * get base D point
	 * 
	 * @return base D point
	 */
	public Point getD() {
		Point p = new Point(xd + modelDx(), yd - modelDy());
		p.translate(-xd, -yd);
		AffineTransform.getRotateInstance(-angle).deltaTransform(p, p);
		p.translate(xd, yd);
		return p;
	}

	/**
	 * move to the specified destination
	 * 
	 * @param nx
	 *            new x coordinate
	 * @param ny
	 *            new y coordinate
	 */
	public void moveTo(int nx, int ny) {
		int xt = nx - xd, yt = ny - yd;
		xd = nx;
		yd = ny;
		baseArea = new Area(AffineTransform.getTranslateInstance((double) xt,
				(double) yt).createTransformedShape(baseArea));
	}

	/**
	 * animate movement
	 * 
	 */
	public void animMovement() {
		if (FB.getInstance() != null && FB.getInstance().isGraphicMode()) {
			action = 'w';
		} else {
			teleport();
		}
	}

	/**
	 * change orientation angle
	 * 
	 * @param newAngle
	 *            the new angle
	 */
	public void setAngle(double newAngle) {
		rotate(newAngle - angle);
	}

	/**
	 * rotate the figure
	 * 
	 * @param deltaAngle
	 *            added angle
	 */
	public void rotate(double deltaAngle) {

		baseArea = new Area(AffineTransform.getRotateInstance(-deltaAngle,
				(double) xd, (double) yd).createTransformedShape(baseArea));
		angle += deltaAngle;
		attribseq();
	}

	/**
	 * rotate figure destination point around specified point
	 * 
	 * @param theta
	 *            rotation anlge
	 * @param xc
	 *            rotation center x coordinate
	 * @param yc
	 *            rotation center y coordinate
	 */
	public void orientate(double theta, int xc, int yc) {

		int xdest = xd;
		int ydest = yd;
		xdest -= xc;
		ydest -= yc;
		Point2D p = AffineTransform.getRotateInstance(-theta).deltaTransform(
				new Point(xdest, ydest), null);
		xdest = (int) p.getX();
		ydest = (int) p.getY();
		xdest += xc;
		ydest += yc;
		moveTo(xdest, ydest);
	}

	/**
	 * change current sprite sequence with specified angle
	 * 
	 * @param angle8
	 *            the 8-base angle
	 */
	protected void setCurrentSpritsequence(int angle8) {
		String name;
		if (action == 's') {
			name = "shoot";
		} else if (action == 't') {
			name = "threat";
		} else if (action == 'w') {
			name = "walk";
		} else if (action == 'd') {
			name = "die";
		} else if (action == 'x') {
			name = "dead";
		} else if (action == 'm') {
			name = "threat";
		} else if (action == 'f') {
			name = "fight";
		} else {
			name = "stand";
		}
		currentSeq = (SpriteSequence) spriteDelegate.getSprite().getSequences()
				.get(name + angle8);

		if (null == currentSeq) {
			currentSeq = (SpriteSequence) spriteDelegate.getSprite()
					.getSequences().get("stand" + angle8);
		}
	}

	/**
	 * get figure bounds considering specified entity scale
	 * 
	 * @param entityScale
	 *            the entity scale
	 * @return the figure bounds
	 */
	public Rectangle getBounds(float entityScale) {

		if (FB.getInstance() == null || FB.getInstance().isBatch()) {
			return new Rectangle();
		}

		int imageWidth;
		int imageHeight;

		imageWidth = spriteDelegate.getSprite().getFrameWitdh();
		imageHeight = spriteDelegate.getSprite().getFrameHeight();

		int rx = (int) (x - imageWidth * entityScale / 2);
		int ry = (int) (y - imageHeight * entityScale / 2);
		int rw = (int) (imageWidth * entityScale);
		int rh = (int) (imageHeight * entityScale);
		return new Rectangle(rx, ry, rw, rh);
	}

	/**
	 * change current frame index
	 * 
	 * @param newFrameIndex
	 *            the new frame index
	 */
	protected void setFrameIndex(int newFrameIndex) {
		if (newFrameIndex != frameIndex) {
			int oldFrameIndex = frameIndex;
			frameIndex = newFrameIndex;
			firePropertyChange(IMAGE, oldFrameIndex, frameIndex);
		}
	}

	/**
	 * apply specified move
	 * 
	 * @param xOffset
	 *            x Offset
	 * @param yOffset
	 *            y Offset
	 */
	public void doMove(int xOffset, int yOffset) {
		if (xOffset != 0 || yOffset != 0) {
			Rectangle previousBounds = getBounds();
			x += xOffset;
			y += yOffset;
			firePropertyChange(BOUNDS, previousBounds, getBounds());
		}
	}

	/**
	 * animate the figure (determine next sprite frame)
	 */
	public void animate() {
		if (visible) {
			if (action == 'w') {
				int distx, disty, pasx, pasy;
				double alpha;
				distx = Math.abs(xd - x);
				disty = Math.abs(yd - y);

				int step = Math.max(spriteDelegate.getStep() / 2, 1);

				if (distx > disty) {
					pasx = (int) (step / (float) (1.0 + (float) disty
							/ (float) distx));
					pasy = step - pasx;
				} else {
					pasy = (int) ((float) (step) / (float) (1.0 + (float) distx
							/ (float) disty));
					pasx = step - pasy;
				}
				if (pasx > distx) {
					pasx = distx;
				}
				if (pasy > disty) {
					pasy = disty;
				}
				if (x > xd) {
					pasx = -pasx;
				}
				if (y > yd) {
					pasy = -pasy;
				}
				alpha = Math.PI + Math.atan2((double) pasx, (double) pasy);
				setCurrentSpritsequence(GeometryTools.to8Base(alpha));
				doMove(pasx, pasy);
				if (x == xd && y == yd) {//end of walk
					action = 'n';
					attribseq();
				}
			}
			if (animationCounter == 1) {
				attribseq();
			}
			if (action == 'd') {
				if (animationCounter == 1) {
					SoundPlayer.getInstance().play("death");
				}
			}
			if (animationCounter == 1 && action == 's') {
				SoundPlayer.getInstance().play("arrow");
			}
			int animindex = java.lang.Math.abs(animationCounter % 4);
			if (animindex == 1) {
				setFrameIndex(currentSeq.getFirstFrameIndex());
			} else if (animindex == 2) {
				setFrameIndex(currentSeq.getSecondFrameIndex());
			} else if (animindex == 3) {
				setFrameIndex(currentSeq.getThirdFrameIndex());
			} else {
				setFrameIndex(currentSeq.getFourthFrameIndex());
			}

			if (action != 'x') {
				animationCounter++;
			}
			if (animationCounter == 5) {
				if (action == 'd') {
					action = 'x';
					animationCounter = 4;
				} else if (action == 's') {
					action = 'n';
					attribseq();
					for (int i = 0; i < projectiles.size(); i++) {
						BattleFieldModel.getInstance().addProjectile(
								(Projectile) projectiles.remove(i));
					}
				} else if (action == 'm') {
					action = 'n';
					attribseq();
					// lancer la magic
					SpellCaster mi = (SpellCaster) spriteDelegate;
					if (!mi.getMagics().isEmpty()) {
						BattleFieldModel.getInstance()
								.addMagics(mi.getMagics());
					}
				}
			}

		} // visible
	}

	/**
	 * paint the figure
	 */
	public void paintImage(Graphics g, int dx, int dy, int scale) {
		if (visible) {
			spriteDelegate.getSprite().drawFrame(g, x, y, scale, dx, dy,
					frameIndex, spriteDelegate.getScale());
		} // visible
	}

	/**
	 * paint the figure on the mini battlefield map
	 * 
	 * @param g
	 *            graphics
	 * @param scaleInverse
	 *            scale factor
	 */
	public void minipaint(Graphics g, float scaleInverse) {
		if (action == 'd') {
			g.setColor(FBColors.background);
		} else {
			g.setColor(spriteDelegate.getColor());
		}
		g.drawLine((int) (x / scaleInverse), (int) (y / scaleInverse),
				(int) (x / scaleInverse), (int) (y / scaleInverse));
	}

	/**
	 * calculate new sprite sequence
	 *
	 */
	public void attribseq() {
		if (FB.getInstance() != null && FB.getInstance().isGraphicMode()) {
			if (action == 's' || action == 't') {
				setCurrentSpritsequence(GeometryTools.to8Base(tmpAngle));
			} else {
				setCurrentSpritsequence(GeometryTools.to8Base(angle));
			}

		}
	}

	/**
	 * animate shooting action
	 * @param shootingAngle shooting angle
	 */
	public void animshoot(double shootingAngle) {
		action = 's';
		animationCounter = -DX.roll(4);
		tmpAngle = shootingAngle;
	}

	/**
	 * get current action
	 * @return current action
	 */
	public char getAction() {
		return action;
	}

	/**
	 * set current action
	 * @param action new action
	 */
	public void setAction(char action) {
		this.action = action;
	}

	/**
	 * get current orientation angle
	 * @return current orientation angle
	 */
	public double getAngle() {
		return angle;
	}

	/**
	 * @return Returns the baseArea.
	 */
	public Area getBaseArea() {
		return baseArea;
	}

	/**
	 * @return Returns the xd.
	 */
	public int getXd() {
		return xd;
	}

	/**
	 * @return Returns the yd.
	 */
	public int getYd() {
		return yd;
	}

	/**
	 * @param animcounter
	 *            The animcounter to set.
	 */
	public void setAnimationCounter(int animcounter) {
		this.animationCounter = animcounter;
	}

	/**
	 * @param xd
	 *            The xd to set.
	 */
	public void setXd(int xd) {
		this.xd = xd;
	}

	/**
	 * @param yd
	 *            The yd to set.
	 */
	public void setYd(int yd) {
		this.yd = yd;
	}

}