/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.campaign;

import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.List;
/**
 * a hexagon
 * used for campaign hexagonal maps
 * @author farfpeuf
 *
 */
public class Hexagon extends Polygon {

	//cosinus(pi/3) constant
	private static final float COS_PI_DIV_3 = 0.5f;
	//sinus(pi/3) constant
	private static final float SIN_PI_DIV_3 = (float) (Math.sqrt(3) / 2.0);

	/**
	 * constructor
	 * @param centerX center x coordinate
	 * @param centerY center y coordinate
	 * @param radius hexagon radius
	 */
	public Hexagon(int centerX, int centerY, int radius) {
		super();
		this.npoints = 6;
		this.xpoints = new int[npoints];
		this.ypoints = new int[npoints];

		int adjacent = (int) (radius * COS_PI_DIV_3);
		int opposed = (int) (radius * SIN_PI_DIV_3);

		xpoints[0] = centerX + radius;
		xpoints[1] = centerX + adjacent;
		xpoints[2] = centerX - adjacent;
		xpoints[3] = centerX - radius;
		xpoints[4] = centerX - adjacent;
		xpoints[5] = centerX + adjacent;

		ypoints[0] = centerY;
		ypoints[1] = centerY + opposed;
		ypoints[2] = centerY + opposed;
		ypoints[3] = centerY;
		ypoints[4] = centerY - opposed;
		ypoints[5] = centerY - opposed;

	}

	/**
	 * get hexagonal coordinates of hexagons surronding a hexagon 
	 * @param hexaCoord the central hexagon coordinates
	 * @param distance the distance to surroundings hexagons
	 * @param bounds to limit search
	 * @return a list of hexagonal coordinates (points)
	 */
	public static List getSurrondingCoords(Point hexaCoord, int distance,
			Rectangle bounds) {
		List result = new ArrayList();
		Point p;

		for (int dist = 1; dist <= distance; dist++) {
			for (int direction = 0; direction < 6; direction++) {
				p = getTranslatedCoords(hexaCoord, dist, direction);
				if (bounds.contains(p)) {
					result.add(p);
				}

				for (int k = 1; k < dist; k++) {
					p = getTranslatedCoords(p, 1, (direction + 2) % 6);
					if (bounds.contains(p)) {
						result.add(p);
					}

				}
			}
		}
		return result;
	}

	/**
	 * get translated hexagonal coordinates of an hexagonal point
	 * @param hexaCoord origin coordinates
	 * @param directDistance translation distance
	 * @param direction translation direction
	 * @return translated hexagonal coordinates
	 */
	public static Point getTranslatedCoords(Point hexaCoord,
			int directDistance, int direction) {
		Point result = new Point(hexaCoord);
		switch (direction) {
		case 0:
			result.translate(0, -directDistance);
			break;
		case 1:
			if (hexaCoord.x % 2 == 0) {
				result.translate(-directDistance, -(directDistance + 1) / 2);
			} else {
				result.translate(-directDistance, -directDistance / 2);
			}
			break;
		case 2:
			if (hexaCoord.x % 2 == 0) {
				result.translate(-directDistance, directDistance / 2);
			} else {
				result.translate(-directDistance, (directDistance + 1) / 2);
			}
			break;
		case 3:
			result.translate(0, directDistance);
			break;
		case 4:
			if (hexaCoord.x % 2 == 0) {
				result.translate(directDistance, directDistance / 2);
			} else {
				result.translate(directDistance, (directDistance + 1) / 2);
			}
			break;
		case 5:
			if (hexaCoord.x % 2 == 0) {
				result.translate(directDistance, -(directDistance + 1) / 2);
			} else {
				result.translate(directDistance, -directDistance / 2);
			}
			break;
		}

		return result;
	}

	/**
	 * get path (list of hexagonal coordinates) from one point to another
	 * @param source source point
	 * @param dest destination point
	 * @return list of hexagonal coordinates
	 */
	public static List getDirectPath(Point source, Point dest) {
		List result = new ArrayList();
		Point p = source;
		int direction;
		while (!p.equals(dest)) {
			direction = getDirectDirection(p, dest);
			p = getTranslatedCoords(p, 1, direction);
			result.add(p);
		}

		return result;
	}

	/**
	 * get distance (number of hexagons to cross) from one point to another
	 * @param source source point
	 * @param dest destination point
	 * @return number of hexagons to cross
	 */
	public static int getDirectDistance(Point source, Point dest) {
		int result = 0;
		Point p = source;
		int direction;
		while (!p.equals(dest)) {
			direction = getDirectDirection(p, dest);
			p = getTranslatedCoords(p, 1, direction);
			result++;
		}

		return result;
	}

	/**
	 * get hexagonal direction (6-base) from one point to another
	 * @param source source point
	 * @param dest destination point
	 * @return the hexagonal direction (6-base)
	 */
	public static int getDirectDirection(Point source, Point dest) {
		if (source.x == dest.x) {
			if (source.y > dest.y) {
				return 0;
			} else {
				return 3;
			}
		} else if (source.x > dest.x) {
			if (source.y < dest.y || (source.y == dest.y && source.x % 2 == 0)) {
				return 2;
			} else {
				return 1;
			}
		} else {
			if (source.y < dest.y || (source.y == dest.y && source.x % 2 == 0)) {
				return 4;
			} else {
				return 5;
			}
		}
	}

}
