/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.main;

import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.geom.Area;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import fb.army.entities.Entity;
import fb.army.hierarchy.Army;
import fb.army.hierarchy.Command;
import fb.battlefield.BattleFieldModel;
import fb.battlefield.MapModel;
import fb.battlefieldelements.VerticalSet;
import fb.battlefieldelements.VerticalSetShadow;
import fb.campaign.Campaign;
import fb.campaign.CampaignGameFrame;
import fb.gui.frames.BattleFrame;
import fb.gui.panels.BattleStatePanel;
import fb.gui.panels.MessageArea;
import fb.phases.BinaryPhase;
import fb.player.PlayersHandler;
import fb.sound.SoundPlayer;
import fb.util.Resources;
import fb.util.StringTools;

/**
 * a battle
 * 
 * @author farfpeuf
 * 
 */
public class Battle {

	// separator used to save campaign battles
	public static char PART_TAG = '$';

	/**
	 * current game map
	 */
	private String currentMap;

	/**
	 * singleton instance
	 */
	private static Battle instance;

	/**
	 * are armies deployed automatically
	 */
	private boolean autoDeploy = false;

	/**
	 * is game played in graphic mode
	 */
	private boolean graphicMode = true;

	/**
	 * current time
	 */
	private int time;

	/**
	 * obstacles
	 */
	private Area lineOfMovementObstacles, lineOfSightObstacles, exterior,
			slowObstacles;

	/**
	 * battlefield area
	 */
	private Rectangle interior;

	/**
	 * involved armies
	 */
	private Army[] armies;

	/**
	 * game main loop
	 */
	private BattleLoop loop;

	/**
	 * involved entities
	 */
	private Vector activeEntities, allEntities;

	/**
	 * players handler
	 */
	private PlayersHandler playersHandler;

	/**
	 * is this battle part of a campaign game
	 */
	private boolean campaignGame;

	/**
	 * was this battle loaded from a file
	 */
	private boolean loadedGame;

	/**
	 * get singleton instance
	 * 
	 * @return the singleton
	 */
	public static synchronized Battle getInstance() {
		if (instance == null) {
			instance = new Battle();
		}
		return instance;
	}

	/**
	 * constructor
	 * 
	 */
	private Battle() {
		// init vecteurs
		campaignGame = false;
		loadedGame = false;
		activeEntities = new Vector();
		allEntities = new Vector();
		playersHandler = new PlayersHandler();
		loop = new BattleLoop();
	}

	/**
	 * build an army
	 * 
	 * @param army
	 *            the army to build
	 * @throws LoadingException
	 */
	public void buildarmy(Army army) throws LoadingException {
		army.build();
		activeEntities.addAll(army.getActiveEntities());
		List groups = army.getAllGroups();
		Command group;
		for (int i = 0; i < groups.size(); i++) {
			group = (Command) groups.get(i);
			allEntities.addAll(group.getAllEntities());
		}

	}

	/**
	 * load involved armies
	 * 
	 * @throws LoadingException
	 */
	public void loadArmies() throws LoadingException {
		try {

			for (int i = 0; i < armies.length; i++) {
				if (armies[i].getList() != null) {
					buildarmy(armies[i]);
				} else {
					activeEntities.addAll(armies[i].getActiveEntities());
					List groups = armies[i].getAllGroups();
					Command group;
					for (int j = 0; j < groups.size(); j++) {
						group = (Command) groups.get(j);
						allEntities.addAll(group.getAllEntities());
					}

				}
			}
			for (int i = 0; i < armies.length; i++) {
				for (int j = i + 1; j < armies.length; j++) {
					if (armies[i].getSide() != armies[j].getSide()) {
						armies[i].addEnnemy(armies[j]);
						armies[j].addEnnemy(armies[i]);
					} else {
						armies[i].addFriend(armies[j]);
						armies[j].addFriend(armies[i]);
					}
				}
			}
		} catch (Exception e) {
			throw new LoadingException(e);
		}
	}

	/**
	 * @return number of active warriors for each army as a string
	 */
	public String getActiveString() {
		String active = null;
		if (armies != null) {

			active = "(";
			Army arm;
			for (int i = 0; i < armies.length; i++) {
				arm = armies[i];
				if (arm != null) {
					active += arm.getNumberOfWarriors();
				}
				if (i < armies.length - 1) {
					active += ",";
				}
			}
			active += ")";
		}
		return (active == null ? "" : active);
	}

	/**
	 * change time period
	 * 
	 */
	public void setTime() {
		if (!FB.getInstance().isBatch()) {
			BattleStatePanel.getInstance().setTime(time);
			if (time <= 5) {
				FB.getInstance().message(
						"-- " + Resources.getString("Newturn") + " - "
								+ Resources.getString("Day") + " (" + time
								+ ") --");
			} else {
				FB.getInstance().message(
						"-- " + Resources.getString("Newturn") + " - "
								+ Resources.getString("Night") + " ("
								+ Integer.toString(time - 5) + ") --");
			}
		}
	}

	/**
	 * change current army
	 * 
	 * @param army
	 *            new army
	 */
	public void setArmy(final Army army) {
		Thread thread = new Thread() {
			public void run() {
				BattleStatePanel.getInstance().setPlayer(army);
			}
		};
		thread.start();
	}

	/**
	 * accelerate time
	 * 
	 * @param plus
	 *            number of added periods
	 */
	public void timePlus(int plus) {
		time += plus;
		if (time > 10) {
			time -= 10;
		}
		if (time < 1) {
			time += 10;
		}
		setTime();
	}

	/**
	 * accelerate time
	 */
	public void timePlus() {
		timePlus(1);
	}

	/**
	 * initialize active entities
	 * 
	 */
	public void initializeModels() {
		Entity entity;
		for (int i = 0; i < activeEntities.size(); i++) {
			entity = (Entity) activeEntities.get(i);
			entity.initBattle();
			entity.create();
		}
	}

	/**
	 * initialize obstacles
	 * 
	 * @param type
	 *            obstacle type
	 * @param area
	 *            obstacles area
	 */
	private void initializeObstacles(char type, Area area) {
		List set = MapModel.getInstance().getSet(type);
		if (set != null) {
			for (int i = 0; i < set.size(); i++) {
				area.add(new Area((Polygon) set.get(i)));
			}
		}
	}

	/**
	 * initialize all obstacles
	 * 
	 */
	public void initializeObstacles() {
		Point vsetPoint;
		lineOfMovementObstacles = new Area();
		slowObstacles = new Area();
		lineOfSightObstacles = new Area();
		initializeObstacles('r', lineOfMovementObstacles);
		initializeObstacles('w', slowObstacles);
		initializeObstacles('w', lineOfSightObstacles);
		initializeObstacles('g', lineOfSightObstacles);
		initializeObstacles('h', lineOfSightObstacles);
		initializeObstacles('h', lineOfMovementObstacles);
		initializeObstacles('m', slowObstacles);
		Enumeration keys = MapModel.getInstance().getVerticalSets().keys();
		String key;
		List vset;
		boolean withShadow;
		while (keys.hasMoreElements()) {
			key = (String) keys.nextElement();
			withShadow = false;
			for (int j = 0; j < MapModel.verticalSetsWithShadows.length; j++) {
				if (MapModel.verticalSetsWithShadows[j].equals(key)) {
					withShadow = true;
				}
			}
			vset = (List) MapModel.getInstance().getVerticalSets().get(key);
			for (int i = 0; i < vset.size(); i++) {
				vsetPoint = (Point) vset.get(i);
				BattleFieldModel.getInstance().addVerticalSet(
						new VerticalSet(vsetPoint, key));

				if (withShadow) {
					BattleFieldModel.getInstance().addVerticalSet(
							new VerticalSetShadow(vsetPoint, key + "S"));
				}

			}
		}
		exterior = new Area(new Rectangle(-500, -500, MapModel.getInstance()
				.getMapWidth() + 1000,
				MapModel.getInstance().getMapLength() + 1000));
		interior = new Rectangle(0, 0, MapModel.getInstance().getMapWidth(),
				MapModel.getInstance().getMapLength());
		exterior.subtract(new Area(interior));
		lineOfSightObstacles.add(new Area(exterior));
		lineOfMovementObstacles.add(new Area(exterior));
	}

	/**
	 * deploy armies
	 * 
	 */
	public void deployArmies() {
		for (int i = 0; i < armies.length; i++) {
			if (loadedGame) {
				armies[i].loadInitialPosition();
			} else {
				List rectangles, ennemyRectangles;
				rectangles = getRectangles(armies[i]);
				ennemyRectangles = getEnnemyRectangles(armies[i]);
				rectangles.add(interior); // for big arms
				armies[i].setInitialPosition(rectangles, ennemyRectangles);
			}
		}
	}

	/**
	 * get ennemies deployment areas for a specified army
	 * 
	 * @param army
	 *            the army
	 * @return ennemies deployment areas
	 */
	public List getEnnemyRectangles(Army army) {
		List result = new ArrayList();
		for (int i = 0; i < armies.length; i++) {
			if (armies[i].getSide() != army.getSide()) {
				if (armies[i].getSide() == 0) {
					result.addAll(MapModel.getInstance().getSet('a'));
				} else if (armies[i].getSide() == 1) {
					result.addAll(MapModel.getInstance().getSet('b'));
				} else if (armies[i].getSide() == 2) {
					result.addAll(MapModel.getInstance().getSet('c'));
				} else if (armies[i].getSide() == 3) {
					result.addAll(MapModel.getInstance().getSet('d'));
				} else if (armies[i].getSide() == 4) {
					result.addAll(MapModel.getInstance().getSet('e'));
				} else {
					result.addAll(MapModel.getInstance().getSet('f'));
				}
			}
		}
		return result;
	}

	/**
	 * get an army deployment area
	 * 
	 * @param army
	 *            the army
	 * @return the deployment area
	 */
	public List getRectangles(Army army) {
		if (army.getSide() == 0) {
			return MapModel.getInstance().getSet('a');
		} else if (army.getSide() == 1) {
			return MapModel.getInstance().getSet('b');
		} else if (army.getSide() == 2) {
			return MapModel.getInstance().getSet('c');
		} else if (army.getSide() == 3) {
			return MapModel.getInstance().getSet('d');
		} else if (army.getSide() == 4) {
			return MapModel.getInstance().getSet('e');
		} else {
			return MapModel.getInstance().getSet('f');
		}
	}

	/**
	 * save battle as a string
	 * 
	 * @return the battle as a string
	 */
	public String saveBattleString() {
		String result = "" + time + "\t" + armies.length + "\t" + currentMap
				+ "\t";

		for (int i = 0; i < armies.length; i++) {
			result += Army.ARM_TAG + armies[i].saveBattleString();
		}
		result += PART_TAG;
		result += ((BinaryPhase) loop.currentSynchronizedPhase).saveString()
				+ "\t";
		for (int i = 0; i < armies.length; i++) {
			result += (loop.getPhaseTokens()[i] ? "y" : "n") + "\t";
		}
		return result;
	}

	/**
	 * save campaign battle as a string
	 * 
	 * @return the campaign battle as a string
	 */
	public String saveCampaignBattleString() {
		String result = "" + time + "\t" + armies.length + "\t" + currentMap
				+ "\t";

		for (int i = 0; i < armies.length; i++) {
			result += Army.ARM_TAG + armies[i].getForce().getId() + "\t"
					+ armies[i].getForceIndex();
		}
		result += PART_TAG;
		result += ((BinaryPhase) loop.currentSynchronizedPhase).saveString()
				+ "\t";
		for (int i = 0; i < armies.length; i++) {
			result += (loop.getPhaseTokens()[i] ? "y" : "n") + "\t";
		}
		return result;
	}
/**
 * load battle part 1 from a stringbuffer read from a file
 * @param buffy the stringbuffer
 * @throws LoadingException
 */
	public void loadBattlePart1(StringBuffer buffy) throws LoadingException {
		try {

			time = StringTools.readBufferInt(buffy);
			armies = new Army[StringTools.readBufferInt(buffy)];
			currentMap = StringTools.readBuffer(buffy);

			String[] tokens = StringTools.readBuffer(buffy, PART_TAG).split(
					Army.ARM_TAG);

			for (int i = 0; i < armies.length; i++) {
				armies[i] = new Army(tokens[i + 1]);
				armies[i].setGmId(i);
			}

		} catch (Exception e) {
			throw new LoadingException(e);
		}
	}
	/**
	 * load campaign battle part 1 from a stringbuffer read from a file
	 * @param buffy the stringbuffer
	 * @throws LoadingException
	 */
	public void loadCampaignBatlePart1(StringBuffer buffy) throws LoadingException {
		try {

			time = StringTools.readBufferInt(buffy);
			armies = new Army[StringTools.readBufferInt(buffy)];
			currentMap = StringTools.readBuffer(buffy);

			String[] tokens = StringTools.readBuffer(buffy, PART_TAG).split(
					Army.ARM_TAG);
			StringBuffer armBuffy;
			for (int i = 0; i < armies.length; i++) {
				armBuffy = new StringBuffer(tokens[i + 1]);
				armies[i] = (Army) Campaign.getInstance().getForces()[StringTools
						.readBufferInt(armBuffy)].getChildAt(StringTools
						.readBufferInt(armBuffy));
				armies[i].setGmId(i);
			}

		} catch (Exception e) {
			throw new LoadingException(e);
		}
	}
	/**
	 * load battle part 2 from a stringbuffer read from a file
	 * @param buffy the stringbuffer
	 * @throws LoadingException
	 */
	public void loadPart2(StringBuffer buffy) throws LoadingException {
		try {

			String className = StringTools.readBuffer(buffy);

			BinaryPhase currentPhase;
			Class currentPhaseClass;
			currentPhaseClass = Class.forName(className);
			currentPhase = (BinaryPhase) currentPhaseClass.newInstance();

			currentPhase.loadString(buffy);

			boolean[] done = new boolean[armies.length];
			int dones = 0;
			for (int i = 0; i < armies.length; i++) {
				done[i] = (StringTools.readBufferChar(buffy) == 'y') ? true
						: false;
				if (done[i]) {
					dones++;
				}
			}

			loop.init();
			loop.setCurrentSynchronizedPhase(currentPhase);
			loop.setPhaseTokens(done);
			loop.setNumberOfPhaseTokens(dones);

		} catch (Exception e) {
			throw new LoadingException(e);
		}
	}
/**
 * get an entity involved in the battle
 * @param id the entity id
 * @return the entity
 */
	public static Entity getEntity(int id) {
		if (id == -1) {
			return null;
		} else {
			return (Entity) getInstance().allEntities.get(id);
		}
	}
/**
 * get a command involved in the battle
 * @param armyId the command army id
 * @param groupId the command id
 * @return the command
 */
	public static Command getGroup(int armyId, int groupId) {
		if (armyId == -1 || groupId == -1) {
			return null;
		} else {
			return (Command) getInstance().armies[armyId].getCommands().get(
					groupId);
		}
	}

	/**
	 * @return battle main loop
	 */
	public static BattleLoop getLoop() {
		return getInstance().loop;
	}
/**
 * stop the battle
 *
 */
	public void stop() {
		try {
			if (Battle.getInstance().graphicMode) {
				BattleFrame.getInstance().stop();
			}
			loop.stop();
		} catch (Exception e) {
			FB.exception(e);
		}
		// reinit vecteurs
		Iterator iter = allEntities.iterator();
		while (iter.hasNext()) {
			((Entity) iter.next()).freeSprite();
		}
		activeEntities.clear();
		allEntities.clear();
		armies = null;
		campaignGame = false;
		loadedGame = false;
		System.gc();
	}


	/**
	 * @return Returns the autoDeploy.
	 */
	public boolean isAutoDeploy() {
		return autoDeploy;
	}

	/**
	 * @param autoDeploy The autoDeploy to set.
	 */
	public void setAutoDeploy(boolean autoDeploy) {
		this.autoDeploy = autoDeploy;
	}

	/**
	 * @return Returns the lineOfMovementObstacles.
	 */
	public Area getLineOfMovementObstacles() {
		return lineOfMovementObstacles;
	}

	/**
	 * @return Returns the lineOfSightObstacles.
	 */
	public Area getLineOfSightObstacles() {
		return lineOfSightObstacles;
	}





	/**
	 * @return Returns the exterior.
	 */
	public Area getExterior() {
		return exterior;
	}

	/**
	 * @return Returns the interior.
	 */
	public Rectangle getInterior() {
		return interior;
	}

	/**
	 * @return Returns the slowObstacles.
	 */
	public Area getSlowObstacles() {
		return slowObstacles;
	}

	/**
	 * @return Returns the time.
	 */
	public int getTime() {
		return time;
	}

	public int winner() {
		List remainingSides = new ArrayList();
		Integer side;
		for (int i = 0; i < armies.length; i++) {
			if (!armies[i].isRoutedAnnihilated()) {
				side = new Integer(armies[i].getSide());
				if (!remainingSides.contains(side)) {
					remainingSides.add(side);
				}
			}
		}
		if (remainingSides.size() > 1) {
			return -1;
		} else if (remainingSides.size() == 1) {
			return ((Integer) remainingSides.get(0)).intValue();
		}
		return -2;
	}
	
	public Army[] getArmies() {
		return armies;
	}

	public void setArmies(Army[] armies) {
		this.armies = armies;
	}

	public Vector getActiveEntities() {
		return activeEntities;
	}

	public Vector getAllEntities() {
		return allEntities;
	}

	public void setTime(int time) {
		this.time = time;
	}

	public String getCurrentMap() {
		return currentMap;
	}

	public void setCurrentMap(String currentMap) {
		this.currentMap = currentMap;
	}

	public PlayersHandler getPlayersHandler() {
		return playersHandler;
	}

	public boolean isCampaignGame() {
		return campaignGame;
	}

	public void setCampaignGame(boolean campaignGame) {
		this.campaignGame = campaignGame;
	}

	public void end(int winner) {

		if (campaignGame) {
			FB.getInstance().clearFrame();
			FB.getInstance().dispalyCampaignGameFrame();
			CampaignGameFrame.getInstance().start();
		} else {
			for (int i = 0; i < getActiveEntities().size(); i++) {
				((Entity) getActiveEntities().get(i)).animThreat();
			}
			SoundPlayer.getInstance().play("threat");
		}

		SoundPlayer.getInstance().play("intro");

		String message;

		if (winner != -2) {
			String victoriousArmies = null;

			for (int i = 0; i < armies.length; i++) {
				if (armies[i].getSide() == winner) {
					if (victoriousArmies == null) {
						victoriousArmies = armies[i].getPlayerName();
					} else {
						victoriousArmies = victoriousArmies + ", "
								+ armies[i].getPlayerName();
					}

				}
			}
			message = Resources.getString("victory");
			message = message.replaceAll("<victorious side>", ""
					+ (char) ('A' + winner));
			message = message.replaceAll("<victorious players>",
					victoriousArmies);
		} else {
			message = Resources.getString("statusquo");
		}
		if (campaignGame) {
			MessageArea.getInstance().clear();
			Campaign.getInstance().endBattle(winner, armies);
			stop();
			FB.getInstance().setDefaultCursor();
		}
		FB.getInstance().message(Resources.getString("endgame"));
		FB.getInstance().popUpMessage(Resources.getString("endgame"), message);

	}

	public void setLoadedGame(boolean loadedGame) {
		this.loadedGame = loadedGame;
	}

}