/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.network;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Vector;

import fb.gui.panels.ChatArea;
import fb.main.FB;
import fb.main.Battle;
import fb.player.Player;
import fb.player.PlayersTableModel;
import fb.util.DX;
import fb.util.FileTools;
import fb.util.Resources;

/**
 * client for network games
 * 
 * @author farfpeuf
 * 
 */
public class Client extends Thread {
	/**
	 * socket used to send and receive network messages
	 */
	private Socket socket;

	/**
	 * writer used to send messages
	 */
	private PrintWriter out;

	/**
	 * reader used to receive messages
	 */
	private BufferedReader in;

	/**
	 * client name
	 */
	private String clientName;

	/**
	 * stop the client ?
	 */
	private boolean stop = false;

	/**
	 * awaited message
	 */
	private String awaited = "";

	/**
	 * synchronization tokens counter
	 */
	private int synchronizer = 0;

	/**
	 * magic synchronization tokens counter
	 */
	private int magicSynchronizer = 0;

	/**
	 * number of clients for this game
	 */
	private int nbClients = 1;

	/**
	 * client id
	 */
	private int id = -1;

	/**
	 * network dice synchronization tokens counter
	 */
	private int netDX = -1;

	/**
	 * number of synchronizations
	 */
	private int synchronizeNB = 0;

	/**
	 * number of magic synchronizations
	 */
	private int magicSynchronizeNB = 0;

	/**
	 * number of messages
	 */
	private int messageNb = 0;

	/**
	 * received Client Lists
	 */
	private Vector receivedClientLists = new Vector();

	/**
	 * wait for network dice roll ?
	 */
	private boolean waitRoll = false;

	/**
	 * singleton instance
	 */
	private static Client instance;

	/**
	 * message sent for synchronization
	 */
	private String synchronizationString;

	/**
	 * get singleton instance
	 * 
	 * @return Returns the singleton.
	 */
	public static Client getInstance() {
		if (instance == null) {
			instance = new Client();
		}
		return instance;
	}

	/**
	 * delete singleton to free memory
	 * 
	 */
	public static synchronized void freeInstance() {
		if (null != instance) {
			instance.destroy();
			instance = null;
		}
	}

	/**
	 * @return true if instance exists
	 */
	public static synchronized boolean hasInstance() {
		return (null != instance);
	}

	/**
	 * constructor initialize thread name
	 */
	private Client() {
		super();
		setName("Client thread");
	}

	/**
	 * @param clientName
	 *            The clientName to set.
	 */
	public void setClientName(String clientName) {
		this.clientName = clientName;
	}

	/**
	 * synchronize with other clients when casting a spell
	 * 
	 * @throws InterruptedException
	 */
	public synchronized void magicSynchronize() throws InterruptedException {
		send(MessagesConstants.MAGICSYNCHRONIZE);
		magicSynchronizeNB++;
		while (magicSynchronizer < nbClients) {
			wait();
		}
		magicSynchronizer -= nbClients;
	}

	/**
	 * roll a network dice
	 * 
	 * @param x
	 *            number of faces
	 * @return dice result
	 */
	public synchronized int dXRoll(int x) {
		int result;
		send(MessagesConstants.ROLL + DX.roll(x));
		if (netDX == -1) {
			waitRoll = true;
			while (netDX == -1) {
				try {
					wait();
				} catch (InterruptedException e) {
					FB.exception(e);
				}
			}
			waitRoll = false;
		}
		result = netDX;
		netDX = -1;
		return result;
	}

	/**
	 * synchronize with other clients
	 */
	public void synchronize() {
		synchronize("");
		synchronize(Battle.getInstance().getActiveString());
	}

	/**
	 * synchronize with other clients
	 * 
	 * @param test
	 *            synchronization test message
	 */
	private synchronized void synchronize(String test) {

		String phaseName = (FB.getInstance() == null || Battle.getLoop()
				.getCurrentPhase() == null) ? "init" : Battle.getLoop()
				.getCurrentPhase().getName();

		synchronizationString = MessagesConstants.SYNCHRONIZE + " "
				+ synchronizeNB + " - " + test + " - " + phaseName;
		send(synchronizationString);
		synchronizeNB++;
		while (synchronizer < nbClients) {
			try {
				wait();
			} catch (InterruptedException e) {
				FB.exception(e);
			}
		}
		synchronizer -= nbClients;
	}

	/**
	 * notify that a synchronization message was received
	 * 
	 * @param notification
	 *            message used to test if client is well synchronized with other
	 *            clients
	 */
	public synchronized void notifySynchronize(String notification) {

		if (synchronizationString != null) {
			if (!synchronizationString.equals(notification)) {
				FB.debug("out of synchronization " + notification + " "
						+ synchronizationString);

				System.exit(-1);
			}
			synchronizationString = null;
		}

		synchronizer++;
		if (synchronizer == nbClients) {
			try {
				notify();
			} catch (IllegalMonitorStateException e) {
				FB.exception(e);
			}
		}
	}

	/**
	 * notify that a magic synchronization message was received
	 */
	public synchronized void notifyMagicSynchronize() {
		magicSynchronizer++;
		if (magicSynchronizer == nbClients) {
			try {
				notify();
			} catch (IllegalMonitorStateException e) {
				FB.exception(e);
			}
		}
	}

	/**
	 * notify that a network dice roll was received
	 * 
	 * @param roll
	 *            the dice score
	 */
	public synchronized void notifyRoll(int roll) {
		if (netDX == -1) {
			netDX = roll;
			if (waitRoll) {
				try {
					notify();
				} catch (IllegalMonitorStateException e) {
					FB.exception(e);
				}
			}
		}
	}

	/**
	 * send a network message
	 * 
	 * @param str
	 *            the message to send
	 */
	public void send(String str) {
		if (out != null && str != null && !str.equals("")) {
			try {
				// FB.debug("SEND " + str);
				out.println(str);
				if (out.checkError())
					throw new IOException();
			} catch (IOException ioe) {
				netError(Resources.getString("lostConnection"));
				FB.exception(ioe);
				try {
					socket.close();
				} catch (IOException ioe2) {
					FB.exception(ioe2);
				}
			}
		}
	}

	/**
	 * send a network message and wait for it
	 * 
	 * @param str
	 *            the message to send
	 */
	public synchronized void sendAndWait(String str) {

		awaited = str;
		send(str);
		try {
			wait();
		} catch (InterruptedException e) {
			FB.exception(e);
		}
	}

	/**
	 * notify that a message was received
	 * 
	 * @param str
	 *            the received message
	 */
	public synchronized void notify(String str) {
		if (str.equals(awaited)) {
			try {
				notify();
			} catch (IllegalMonitorStateException e) {
				FB.exception(e);
			}
		}
	}

	/**
	 * handle network error
	 * 
	 * @param message
	 *            the error message
	 */
	private void netError(String message) {

		ChatArea.getInstance().writeText(message, new Color(255, 1, 1));

		FB.getInstance().popUpMessage(Resources.getString("netError"), message);
	}

	/**
	 * connect client to server
	 * 
	 * @param hostname
	 *            server name
	 * @param port
	 *            connection port
	 * @return true if connection is successfull
	 */
	public boolean connect(String hostname, int port) {
		InetAddress host = null;
		try {
			host = InetAddress.getByName(hostname);
		} catch (UnknownHostException e) {
			netError(Resources.getString("unknownHost") + hostname);
			return false;
		} catch (NumberFormatException nfe) {
			netError(Resources.getString("invalidPort"));
			return false;
		}
		// connection
		try {
			socket = new Socket(host, port);
		} catch (IOException ioe1) {
			netError(Resources.getString("unableToConnect") + hostname);
			return false;
		}
		// ----------
		try {
			out = new PrintWriter(socket.getOutputStream(), true);
		} catch (IOException e) {
			netError(Resources.getString("unableToGetOutStream") + hostname);

			try {
				socket.close();
			} catch (IOException ioe) {
				FB.exception(ioe);
			}
			return false;
		}
		try {
			in = new BufferedReader(new InputStreamReader(socket
					.getInputStream()));
			if (in.readLine().equals("GAMEFULL")) {
				netError(Resources.getString("gameFull") + hostname);
				try {
					out.close();
					in.close();
					socket.close();
				} catch (IOException ioe2) {
					FB.exception(ioe2);
				}
				return false;
			}

		} catch (IOException e) {
			netError(Resources.getString("unableToGetInStream") + hostname);
			out.close();
			try {
				socket.close();
			} catch (IOException ioe) {
				FB.exception(ioe);
			}
			return false;
		}
		return true;
	}

	/**
	 * client loop to receive messages
	 */
	public void run() {
		String msg = null;
		try {
			while (!stop) {
				msg = in.readLine();
				messageNb++;
				if (msg != null) {
					MessagesInterpreter.interprete(msg);
					notify(msg);
					yield();
				} else {
					break;
				}
			}
		} catch (IOException lostConnection) {
			netError(Resources.getString("lostConnection"));
			FB.exception(lostConnection);
		} finally {
			try {
				socket.close();
			} catch (IOException ioe) {
				FB.exception(ioe);
			}
		}
	}

	/**
	 * destroy the client
	 */
	public void destroy() {
		stop = true;
	}

	/**
	 * Returns the playerName.
	 * 
	 * @return String
	 */
	public String getClientName() {
		return clientName;
	}

	/**
	 * send client armies
	 * 
	 */
	public void sendArmies() {
		String filestream, message = MessagesConstants.MYARMIESARE + id + "#";
		PlayersTableModel localPlas = Battle.getInstance().getPlayersHandler()
				.getLocalPlayersTableModel();
		Player pla;
		for (int i = 0; i < localPlas.getNbPlayer(); i++) {
			pla = localPlas.getPla(i);
			filestream = FileTools.readFile("armies" + File.separator
					+ pla.getArmyName());
			filestream = filestream.replaceAll("\n", "@");
			message += pla.getName() + "#" + pla.isIa() + "#" + pla.getSide()
					+ "#" + filestream + "#";
		}
		send(message);
	}

	/**
	 * send client players
	 * 
	 */
	public void sendPlayers() {
		String message = MessagesConstants.MYPLAYERSARE + id + "#" + clientName
				+ "#";
		PlayersTableModel localPlayers = Battle.getInstance()
				.getPlayersHandler().getLocalPlayersTableModel();
		Player pla;
		for (int i = 0; i < localPlayers.getNbPlayer(); i++) {
			pla = localPlayers.getPla(i);
			message += pla.getName() + "#" + pla.getArmyName() + "#"
					+ pla.isIa() + "#" + pla.getSide() + "#";
		}
		send(message);

	}

	/**
	 * Returns the id.
	 * 
	 * @return int
	 */
	public int getClientId() {
		return id;
	}

	/**
	 * Sets the id.
	 * 
	 * @param id
	 *            The id to set
	 */
	public void setId(int id) {
		this.id = id;
	}

	/**
	 * test if a client list was received
	 * 
	 * @param id
	 *            the client id
	 * @return true if the client list was received
	 */
	public boolean hasReceivedClientLists(int id) {
		if (receivedClientLists.contains(new Integer(id))) {
			return true;
		}
		receivedClientLists.add(new Integer(id));
		return false;
	}

	/**
	 * @param nbClients
	 *            The nbClients to set.
	 */
	public void setNbClients(int nbClients) {
		this.nbClients = nbClients;
	}

}