/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.util;

import java.awt.Graphics;
import java.awt.Image;

import javax.swing.ImageIcon;

/**
 * a graphical image with custom methods designed for fantasy battlefields
 * 
 * @author farfpeuf
 * 
 */
public class FBImage {

	/**
	 * number of point of views for sprite images
	 */
	private static final int NB_POINT_OF_VIEW = 8;

	/**
	 * delegate java image
	 */
	private Image image;

	/**
	 * sprite frames size
	 */
	private int framePixelWidth, framePixelHeight;

	/**
	 * image width and height for game
	 */
	private int gameWidth, gameHeight;

	/**
	 * constructor using image icon for non-sprites
	 * 
	 * @param icon
	 *            the icon
	 */
	public FBImage(ImageIcon icon) {
		this.image = icon.getImage();
		gameWidth = icon.getIconWidth() / 8;
		gameHeight = icon.getIconHeight() / 8;
	}

	/**
	 * constructor using image icon for sprites
	 * 
	 * @param icon
	 *            the icon
	 * @param numberOfFrames
	 *            the sprite number of frames
	 */
	public FBImage(ImageIcon icon, int numberOfFrames) {
		image = icon.getImage();
		framePixelWidth = icon.getIconWidth()
				/ (numberOfFrames / NB_POINT_OF_VIEW);
		framePixelHeight = icon.getIconHeight() / NB_POINT_OF_VIEW;
		gameWidth = framePixelWidth / 8;
		gameHeight = framePixelHeight / 8;
	}

	/**
	 * draw the whole image
	 * 
	 * @param g
	 *            graphics
	 * @param x
	 *            center x coordinate
	 * @param y
	 *            center y coordinate
	 * @param scale
	 *            scale
	 * @param dx
	 *            x offset
	 * @param dy
	 *            y offset
	 */
	public void draw(Graphics g, int x, int y, float scale, int dx, int dy) {
		int destinationX = (int) ((x - gameWidth / 2) * scale + dx), destinationY = (int) ((y - gameHeight / 2)
				* scale + dy), destinationWidth = (int) (gameWidth * scale), destinationHeight = (int) (gameHeight * scale);
		g.drawImage(image, destinationX, destinationY, destinationWidth,
				destinationHeight, null);
	}

	/**
	 * draw a frame of the image
	 * 
	 * @param g
	 *            graphics
	 * @param x
	 *            center x coordinate
	 * @param y
	 *            center y coordinate
	 * @param scale
	 *            game scale
	 * @param dx
	 *            x offset
	 * @param dy
	 *            y offset
	 * @param index
	 *            frame index
	 * @param entityscale
	 *            entity scale
	 */
	public void drawFrame(Graphics g, int x, int y, float scale, int dx,
			int dy, int index, float entityscale) {
		int destinationX1 = (int) ((x - gameWidth * entityscale / 2) * scale + dx), destinationY1 = (int) ((y - gameHeight
				* entityscale / 2)
				* scale + dy), sourceX1 = (int) (index / 8) * framePixelWidth, sourceY1 = (int) (index % 8)
				* framePixelHeight,

		destinationX2 = (int) (destinationX1 + (gameWidth * entityscale * scale)), destinationY2 = (int) (destinationY1 + (gameHeight
				* entityscale * scale)), sourceX2 = sourceX1 + framePixelWidth, sourceY2 = sourceY1
				+ framePixelHeight;

		g.drawImage(image, destinationX1, destinationY1, destinationX2,
				destinationY2, sourceX1, sourceY1, sourceX2, sourceY2, null);
	}

	/**
	 * Returns the height.
	 * 
	 * @return int
	 */
	public int getGameHeight() {
		return gameHeight;
	}

	/**
	 * Returns the image.
	 * 
	 * @return Image
	 */
	public Image getImage() {
		return image;
	}

	/**
	 * Returns the width.
	 * 
	 * @return int
	 */
	public int getGameWidth() {
		return gameWidth;
	}

}
