/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.util;

import java.awt.Point;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.PathIterator;

/**
 * utility class for geometry
 * 
 * @author farfpeuf
 * 
 */
public class GeometryTools {
	/**
	 * pi/4 constant
	 */
	private static final double PI_DIV_4 = Math.PI / 4.0;

	/**
	 * rotate a point around another one
	 * 
	 * @param theta
	 *            rotation angle
	 * @param p
	 *            the point to rotate
	 * @param xc
	 *            rotation center x coordinate
	 * @param yc
	 *            rotation center y coordinate
	 */
	public static void rotatePoint(double theta, Point p, int xc, int yc) {
		p.translate(-xc, -yc);
		AffineTransform.getRotateInstance(-theta).deltaTransform(p, p);
		p.translate(xc, yc);
	}

	/**
	 * calculate transformed area after a rotation and a translaation
	 * 
	 * @param area
	 *            the area
	 * @param srcX
	 *            initial x coordinate
	 * @param srcY
	 *            initial y coordinate
	 * @param destX
	 *            destination x coordinate
	 * @param destY
	 *            destination y coordinate
	 * @param theta
	 *            rotation angle
	 * @return the transformed area
	 */
	public static Area transformedArea(Area area, int srcX, int srcY,
			int destX, int destY, double theta) {
		Area result = new Area(area);

		result.transform(AffineTransform.getRotateInstance(theta, srcX, srcY));

		result.transform(AffineTransform.getTranslateInstance(destX - srcX,
				destY - srcY));

		return result;
	}

	/**
	 * calculate translated area
	 * 
	 * @param area
	 *            initial area
	 * @param dx
	 *            x offset
	 * @param dy
	 *            y offset
	 * @return translated area
	 */
	public static Area translatedArea(Area area, int dx, int dy) {
		return new Area(AffineTransform.getTranslateInstance((double) dx,
				(double) dy).createTransformedShape(area));
	}

	/**
	 * calculate rounded angle between a vector and the x axis
	 * 
	 * @param srcX
	 *            vector origin x coordinate
	 * @param srcY
	 *            vector origin y coordinate
	 * @param destX
	 *            vector destination x coordinate
	 * @param destY
	 *            vector destination y coordinate
	 * @return rounded angle between vector and the x axis
	 */
	public static double getRoundedAngle(int srcX, int srcY, int destX,
			int destY) {
		return rounded(getAngle(srcX, srcY, destX, destY));
	}

	/**
	 * calculate angle between a vector and the x axis
	 * 
	 * @param srcX
	 *            vector origin x coordinate
	 * @param srcY
	 *            vector origin y coordinate
	 * @param destX
	 *            vector destination x coordinate
	 * @param destY
	 *            vector destination y coordinate
	 * @return angle between vector and the x axis
	 */
	public static double getAngle(int srcX, int srcY, int destX, int destY) {
		int distX = destX - srcX;
		int distY = destY - srcY;
		double deupi = 2 * Math.PI, theta;
		theta = -Math.atan2(distY, distX);
		theta -= Math.PI / 2;
		return theta % deupi;
	}

	/**
	 * converts 8-Base Angle to radian angle
	 * 
	 * @param _8BaseAngle
	 *            the 8-Base Angle
	 * @return the radian angle
	 */
	public static double angleRadian(int _8BaseAngle) {
		if (_8BaseAngle <= 4) {
			return PI_DIV_4 * _8BaseAngle;
		} else {
			return -PI_DIV_4 * (8 - _8BaseAngle);
		}
	}

	/**
	 * rounds angle to closest pi/4 multiple
	 * 
	 * @param theta
	 *            the angle
	 * @return the rounded angle
	 */
	public static double rounded(double theta) {
		return rounded(theta, PI_DIV_4);
	}

	/**
	 * rounds angle to closest round multiple
	 * 
	 * @param theta
	 *            the angle
	 * @param round
	 *            the round
	 * @return the rounded angle
	 */
	public static double rounded(double theta, double round) {
		int nbTr = (int) (theta / round);
		return nbTr * round;
	}

	/**
	 * converts radian angle to 8-Base Angle
	 * 
	 * @param theta
	 *            the radian Angle
	 * @return the 8-Base angle
	 */
	public static int to8Base(double theta) {
		double deupi = 2 * Math.PI;
		theta = theta % deupi;
		if (theta < -5.8905) {
			return 0;
		} else if (theta < -5.1051) {
			return 1;
		} else if (theta < -4.3197) {
			return 2;
		} else if (theta < -3.5343) {
			return 3;
		} else if (theta < -2.7489) {
			return 4;
		} else if (theta < -1.9635) {
			return 5;
		} else if (theta < -1.1781) {
			return 6;
		} else if (theta < -0.3927) {
			return 7;
		} else if (theta < 0.3927) {
			return 0;
		} else if (theta < 1.1781) {
			return 1;
		} else if (theta < 1.9635) {
			return 2;
		} else if (theta < 2.7489) {
			return 3;
		} else if (theta < 3.5343) {
			return 4;
		} else if (theta < 4.3197) {
			return 5;
		} else if (theta < 5.1051) {
			return 6;
		} else if (theta < 5.8905) {
			return 7;
		} else {
			return 0;
		}
	}

	/**
	 * calculates if a segment has an intersection with an area
	 * 
	 * @param S1P0
	 *            segment first point
	 * @param S1P1
	 *            segment second point
	 * @param polygonalArea
	 *            the area
	 * @return true if the segment has an intersection with the area
	 */
	public static boolean hasIntersection(Point S1P0, Point S1P1,
			Area polygonalArea) {

		if (polygonalArea.contains(S1P0) || polygonalArea.contains(S1P1)) {
			return true;
		}

		PathIterator path = polygonalArea.getPathIterator(null);
		float[] coords = new float[6];
		float curx = 0, cury = 0;
		float movx = 0, movy = 0;
		float newx, newy;
		while (!path.isDone()) {
			switch (path.currentSegment(coords)) {
			case PathIterator.SEG_MOVETO:
				movx = curx = coords[0];
				movy = cury = coords[1];
				break;
			case PathIterator.SEG_LINETO:
				newx = coords[0];
				newy = coords[1];
				if (hasIntersection(S1P0, S1P1, new Point((int) curx,
						(int) cury), new Point((int) newx, (int) newy))) {
					return true;
				}
				curx = newx;
				cury = newy;
				break;
			case PathIterator.SEG_CLOSE:
				if (hasIntersection(S1P0, S1P1, new Point((int) movx,
						(int) movy), new Point((int) curx, (int) cury))) {
					return true;
				}
				curx = movx;
				cury = movy;
				break;
			}
			path.next();
		}
		return false;

	}

	/**
	 * calculates if 2 segments have an intersection
	 * 
	 * @param S1P0
	 *            segment 1 first point
	 * @param S1P1
	 *            segment 1 second point
	 * @param S2P0
	 *            segment 2 first point
	 * @param S2P1
	 *            segment 2 second point
	 * @return true if the 2 segments have an intersection
	 */
	public static boolean hasIntersection(Point S1P0, Point S1P1, Point S2P0,
			Point S2P1) {
		return segmentsIntersection(S1P0, S1P1, S2P0, S2P1) != null;
	}

	/**
	 * perp product between 2 vectors
	 * 
	 * @param u
	 *            first vector
	 * @param v
	 *            second vector
	 * @return the perp product
	 */
	public static float perp(Point u, Point v) {
		return u.x * v.y - u.y * v.x;
	}

	/**
	 * dot product between 2 vectors
	 * 
	 * @param u
	 *            first vector
	 * @param v
	 *            second vector
	 * @return the dot product
	 */
	public static float dot(Point u, Point v) {
		return u.x * v.x + u.y * v.y;
	}

	/**
	 * calculates intersection between 2 segments
	 * 
	 * @param S1P0
	 *            segment 1 first point
	 * @param S1P1
	 *            segment 1 second point
	 * @param S2P0
	 *            segment 2 first point
	 * @param S2P1
	 *            segment 2 second point
	 * @return intersection points 0=disjoint (no intersect) 1=intersection in
	 *         unique point I0 2=overlap in segment from I0 to I1
	 */

	public static Point[] segmentsIntersection(Point S1P0, Point S1P1,
			Point S2P0, Point S2P1) {

		Point u = new Point(S1P1.x - S1P0.x, S1P1.y - S1P0.y);
		Point v = new Point(S2P1.x - S2P0.x, S2P1.y - S2P0.y);
		Point w = new Point(S1P0.x - S2P0.x, S1P0.y - S2P0.y);
		float D = perp(u, v);

		// test if they are parallel (includes either being a point)
		if (Math.abs(D) < 0.00000001) { // S1 and S2 are parallel
			if (perp(u, w) != 0 || perp(v, w) != 0) {
				return null; // they are NOT collinear
			}
			// they are collinear or degenerate
			// check if they are degenerate points
			float du = dot(u, u);
			float dv = dot(v, v);
			if (du == 0 && dv == 0) { // both segments are points
				if (S1P0.equals(S2P0)) // they are distinct points
					return null;
				// they are the same point
				return new Point[] { new Point(S1P0) };
			}
			if (du == 0) { // S1 is a single point
				if (inSegment(S1P0, S2P0, S2P1) == 0) // but is not in S2
					return null;

				return new Point[] { new Point(S1P0) };
			}
			if (dv == 0) { // S2 a single point
				if (inSegment(S2P0, S1P0, S1P1) == 0) // but is not in S1
					return null;
				return new Point[] { new Point(S2P0) };
			}
			// they are collinear segments - get overlap (or not)
			float t0, t1; // endpoints of S1 in eqn for S2
			Point w2 = new Point(S1P1.x - S2P0.x, S1P1.y - S2P0.y);
			if (v.x != 0) {
				t0 = w.x / v.x;
				t1 = w2.x / v.x;
			} else {
				t0 = w.y / v.y;
				t1 = w2.y / v.y;
			}
			if (t0 > t1) { // must have t0 smaller than t1
				float t = t0;
				t0 = t1;
				t1 = t; // swap if not
			}
			if (t0 > 1 || t1 < 0) {
				return null; // NO overlap
			}
			t0 = t0 < 0 ? 0 : t0; // clip to min 0
			t1 = t1 > 1 ? 1 : t1; // clip to max 1
			if (t0 == t1) { // intersect is a point
				return new Point[] { new Point((int) (S2P0.x + t0 * v.x),
						(int) (S2P0.y + t0 * v.y)) };
			}

			// they overlap in a valid subsegment
			return new Point[] {
					new Point((int) (S2P0.x + t0 * v.x), (int) (S2P0.y + t0
							* v.y)),
					new Point((int) (S2P0.x + t1 * v.x), (int) (S2P0.y + t1
							* v.y)) };
		}

		// the segments are skew and may intersect in a point
		// get the intersect parameter for S1
		float sI = perp(v, w) / D;
		if (sI < 0 || sI > 1) // no intersect with S1
			return null;

		// get the intersect parameter for S2
		float tI = perp(u, w) / D;
		if (tI < 0 || tI > 1) // no intersect with S2
			return null;
		// compute S1 intersect point
		return new Point[] { new Point((int) (S1P0.x + sI * u.x),
				(int) (S1P0.y + sI * u.y)) };
	}

	/**
	 * determine if a point is inside a segment
	 * 
	 * @param P
	 *            a point
	 * @param SP0
	 *            segment first point
	 * @param SP1
	 *            segment second point
	 * @return 1 = P is inside S 0 = P is not inside S
	 */
	public static int inSegment(Point P, Point SP0, Point SP1) {
		if (SP0.x != SP1.x) { // S is not vertical
			if (SP0.x <= P.x && P.x <= SP1.x)
				return 1;
			if (SP0.x >= P.x && P.x >= SP1.x)
				return 1;
		} else { // S is vertical, so test y coordinate
			if (SP0.y <= P.y && P.y <= SP1.y)
				return 1;
			if (SP0.y >= P.y && P.y >= SP1.y)
				return 1;
		}
		return 0;
	}
	
	
	/**
	 * calculate the surface of a ploygonal area
	 * @param polygonalArea the area
	 * @return the surface
	 */
	public static int calculateSurface(Area polygonalArea) {

		int surface = 0;
		PathIterator path = polygonalArea.getPathIterator(null);
		float[] coords=new float[6];
		float curx = 0, cury = 0;
		float movx = 0, movy = 0;
		float newx, newy;
		while(!path.isDone()){
			switch (path.currentSegment(coords)) {
			case PathIterator.SEG_MOVETO:
			movx = curx = coords[0];
			movy = cury = coords[1];
			break;
			case PathIterator.SEG_LINETO:
			newx = coords[0];
			newy = coords[1];
			surface += (int)((newx - curx) * (cury+ newy));
			curx = newx;
			cury = newy;
			break;
			case PathIterator.SEG_CLOSE:
			surface += (int)((movx - curx) * (cury+ movy));
			curx = movx;
			cury = movy;
			break;
			}
			path.next();
		}
		return surface;

	}
	

}