/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.util;

import java.awt.Dimension;
import java.awt.Image;
import java.awt.Toolkit;
import java.io.File;
import java.util.Hashtable;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import javax.swing.ImageIcon;

import fb.main.FB;
import fb.main.Options;

/**
 * fantasy battlefields resources manager
 * 
 * @author farfpeuf
 */
public class Resources {
	/**
	 * delegate resource bundle
	 */
	private static ResourceBundle resourceBundle;

	/**
	 * images cache table
	 */
	private static Hashtable imageTable = new Hashtable(), oldTable;

	/**
	 * icon caches
	 */
	private static ImageIcon[] phaseicons, timeicons;

	/**
	 * load phases icons into memory
	 * 
	 */
	public static void loadPhasesIcons() {
		if (phaseicons == null) {
			phaseicons = new ImageIcon[] { getSign("threatphase"),
					getSign("shootingphase"), getSign("movementphase"),
					getSign("combatphase"), getSign("newordersphase"),
					getSign("magicphase") };
		}
	}

	/**
	 * load time icons into memory
	 * 
	 */
	public static void loadTimeIcons() {
		if (timeicons == null) {
			timeicons = new ImageIcon[] { getSign("time1"), getSign("time2"),
					getSign("time3"), getSign("time4"), getSign("time5"),
					getSign("time7"), getSign("time8"), getSign("time9"),
					getSign("time10"), getSign("time11"), };
		}
	}

	/**
	 * get a time icon
	 * 
	 * @param index
	 *            icon index
	 * @return the time icon
	 */
	public static ImageIcon getTimeIcon(int index) {
		return timeicons[index];
	}

	/**
	 * get a phase icon
	 * 
	 * @param index
	 *            phase index
	 * @return the phase icon
	 */
	public static ImageIcon getPhaseIcon(int index) {
		return phaseicons[index];
	}

	/**
	 * delete image table and prepare to free memory
	 */
	public static void free() {
		oldTable = imageTable;
		imageTable = new Hashtable();
	}

	/**
	 * delete an image from memory
	 * 
	 * @param key
	 *            the image key
	 */
	public static void free(String key) {
		imageTable.remove(key);
	}

	/**
	 * get a message from localized bundle
	 * 
	 * @param key
	 *            the message key
	 * @return the localized message
	 */
	public static String getString(String key) {
		if (null != resourceBundle) {
			try {
				return resourceBundle.getString(key);
			} catch (MissingResourceException e) {
				FB.debug("Missing value for " + key);
				return key;
			}
		} else {
			return key;
		}
	}

	/**
	 * get an image
	 * 
	 * @param key
	 *            the image key
	 * @return the image corresponding to key
	 */
	private static ImageIcon getImageIcon(String key) {
		ImageIcon icon = (ImageIcon) imageTable.get(key);
		if (null == icon) {
			if (null != oldTable) {
				icon = (ImageIcon) oldTable.get(key);
			}
			if (null == icon) {

				icon = FileTools.getIcon(key);

			}
			imageTable.put(key, icon);
		}
		return icon;
	}

	/**
	 * get an image
	 * 
	 * @param key
	 *            the image key
	 * @return the image corresponding to key or a empty image if not found
	 */
	public static ImageIcon getImage(String key) {
		ImageIcon icon = getImageIcon(key);
		if (null == icon) {
			icon = new ImageIcon();
		}
		return icon;
	}

	/**
	 * get a button image
	 * 
	 * @param name
	 *            the button name
	 * @return the button image
	 */
	public static ImageIcon getButtonIcon(String name) {
		return getImage("buttons" + File.separator + name + ".gif");
	}

	/**
	 * get a image map key
	 * 
	 * @param name
	 *            the map name
	 * @param flipLR
	 *            is map flipped from left to right ?
	 * @param flipUD
	 *            is map flipped from up to down ?
	 * @return the map key
	 */
	public static String getMapKey(String name, boolean flipLR, boolean flipUD) {
		String path = "maps" + File.separator + name + ".jpg";
		return path + "," + flipLR + "," + flipUD;
	}

	/**
	 * get a map image
	 * 
	 * @param name
	 *            the map name
	 * @param flipLR
	 *            is map flipped from left to right ?
	 * @param flipUD
	 *            is map flipped from up to down ?
	 * @return the map image
	 */
	public static Image getMap(String name, boolean flipLR, boolean flipUD) {

		String path = "maps" + File.separator + name;
		String key = path + "," + flipLR + "," + flipUD;

		Image flippedImage = (Image) imageTable.get(key);
		if (null == flippedImage) {
			if (null != oldTable) {
				flippedImage = (Image) oldTable.get(key);
			}
			if (null == flippedImage) {
				ImageIcon nonFlipped;

				nonFlipped = FileTools.getIcon(path);

				flippedImage = ImageTools.flip(nonFlipped, flipLR, flipUD);
			}
			imageTable.put(key, flippedImage);
		}
		return flippedImage;

	}

	/**
	 * get a tile transition image
	 * 
	 * @param direction
	 *            the transition initial direction
	 * @param length
	 *            the transition length
	 * @param layer
	 *            the transition layer
	 * @return the tile transition image
	 */
	public static Image getTileTransition(int direction, int length, int layer) {
		String type = "a";
		boolean flipLR = false;
		boolean flipUD = false;
		if (length == 1) {
			switch (direction) {
			case 0:
				break;
			case 1:
				type = "b";
				break;
			case 2:
				type = "b";
				flipUD = true;
				break;
			case 3:
				flipUD = true;
				break;
			case 4:
				type = "b";
				flipUD = true;
				flipLR = true;
				break;
			default:
				type = "b";
				flipLR = true;
				break;
			}
		} else if (length == 2) {
			switch (direction) {
			case 0:
				break;
			case 1:
				type = "b";
				break;
			case 2:
				flipUD = true;
				break;
			case 3:
				flipUD = true;
				flipLR = true;
				break;
			case 4:
				type = "b";
				flipLR = true;
				break;
			default:
				flipLR = true;
				break;
			}
		} else if (length == 3) {
			switch (direction) {
			case 0:
				break;
			case 1:
				flipUD = true;
				break;
			case 2:
				type = "b";
				;
				break;
			case 3:
				flipUD = true;
				flipLR = true;
				break;
			case 4:
				flipLR = true;
				break;
			default:
				type = "b";
				flipUD = true;
				break;
			}
		}

		String path = "tile" + File.separator + "T" + layer + "-" + length
				+ "-" + type + ".png";
		String key = path + "," + direction;

		Image rotatedImage = (Image) imageTable.get(key);
		if (null == rotatedImage) {
			if (null != oldTable) {
				rotatedImage = (Image) oldTable.get(key);
			}
			if (null == rotatedImage) {
				ImageIcon nonRotated;

				nonRotated = FileTools.getIcon(path);

				// rotatedImage = ImageTools.rotate(nonRotated,
				// -((double)direction)*Math.PI/3.0);

				rotatedImage = ImageTools.flip(nonRotated, flipLR, flipUD);
			}
			imageTable.put(key, rotatedImage);
		}
		return rotatedImage;
	}

	/**
	 * get a cursor image
	 * 
	 * @param name
	 *            the cursor name
	 * @return the cursor image
	 */
	public static ImageIcon getCursor(String name) {
		return getImageIcon("cursors" + File.separator + name + ".gif");
	}

	/**
	 * get a look and feel image
	 * 
	 * @param name
	 *            the look and feel name
	 * @return the look and feel image
	 */
	public static ImageIcon getLookNFeelImage(String name) {
		return getImage("look" + File.separator + name + ".jpg");
	}

	/**
	 * get a sign image
	 * 
	 * @param name
	 *            the sign name
	 * @return the sign image
	 */
	public static ImageIcon getSign(String name) {
		return getImage("signs" + File.separator + name + ".gif");
	}

	/**
	 * get a mini-icon image
	 * 
	 * @param name
	 *            the mini-icon name
	 * @return the mini-icon image
	 */
	public static ImageIcon getMinicon(String name) {
		return getScaledImage("signs" + File.separator + name + ".gif", 20, 20);
	}

	/**
	 * get a vertical set image
	 * 
	 * @param name
	 *            the vertical set name
	 * @return the vertical set image
	 */
	public static ImageIcon getVerticalSetImage(String name) {
		return getImage("vset" + File.separator + name + ".gif");
	}

	/**
	 * get a tile image
	 * 
	 * @param name
	 *            the tile name
	 * @return the tile image
	 */
	public static ImageIcon getTileImage(char name) {
		return getImage("tile" + File.separator + name + ".png");
	}

	/**
	 * get a scaled image
	 * 
	 * @param path
	 *            the image path
	 * @param width
	 *            the image width
	 * @param height
	 *            the image length
	 * @return the scaled image
	 */
	public static ImageIcon getScaledImage(String path, int width, int height) {
		String key = path + "," + width + "," + height;
		ImageIcon scaledIcon = (ImageIcon) imageTable.get(key);
		if (null == scaledIcon) {
			if (null != oldTable) {
				scaledIcon = (ImageIcon) oldTable.get(key);
			}
			if (null == scaledIcon) {
				ImageIcon nonScaledIcon = getImage(path);
				scaledIcon = new ImageIcon((nonScaledIcon).getImage()
						.getScaledInstance(width, height, Image.SCALE_SMOOTH));
			}
			imageTable.put(key, scaledIcon);
		}
		return scaledIcon;
	}

	/**
	 * get a background image
	 * 
	 * @param name
	 *            the image name
	 * @return the background image
	 */
	public static ImageIcon getBackgroundImage(String name) {
		String key = "look" + File.separator + name + ".jpg";
		Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
		int width = (int) screenDim.getWidth();
		int height = (int) screenDim.getHeight();
		ImageIcon scaledIcon = (ImageIcon) imageTable.get(key);
		if (null == scaledIcon) {
			if (null != oldTable) {
				scaledIcon = (ImageIcon) oldTable.get(key);
			}
			if (null == scaledIcon) {
				ImageIcon nonScaledIcon;

				nonScaledIcon = FileTools.getIcon(key);

				scaledIcon = new ImageIcon((nonScaledIcon).getImage()
						.getScaledInstance(width, height, Image.SCALE_FAST));
			}
			imageTable.put(key, scaledIcon);
		}
		return scaledIcon;
	}

	/**
	 * @return localized html manual
	 */
	public static String getManual() {
		String manualPath;

		if (Options.getLanguage().equals("fr")) {
			manualPath = "manuel_de_jeu.htm";
		} else {
			manualPath = "game_manual.htm";
		}

		return FileTools.readFile(manualPath, false);
	}

	/**
	 * set the resource bundle for localized strings
	 * 
	 * @param resourceBundle
	 *            the new resource bundle
	 */
	public static void setResourceBundle(ResourceBundle resourceBundle) {
		Resources.resourceBundle = resourceBundle;
	}
}